import { lcaIcon } from './icons';
import { ToolbarButton } from '@jupyterlab/apputils';
import { initMixpanel } from './mixpanel';
import { INotebookTracker, NotebookActions } from "@jupyterlab/notebook";
import { toArray, toObject } from '@lumino/algorithm';
import '../style/style.css';
/**
 * The command IDs used by the plugin.
 */
export var CommandIDs;
(function (CommandIDs) {
    CommandIDs.lcaInit = 'notebook:init-low-code-assistant';
})(CommandIDs || (CommandIDs = {}));
class LCAInitButton {
    constructor(commands, notebookTracker) {
        this._commands = commands;
        this._notebookTracker = notebookTracker;
    }
    createNew(panel) {
        const nb = this._notebookTracker;
        const button = new ToolbarButton({
            className: 'lcaInit',
            tooltip: 'Initialize Low Code Assistant™',
            label: "Low Code Assistant™",
            icon: lcaIcon,
            onClick: () => {
                const current = panel;
                NotebookActions.insertAbove(current.content);
                const activeCell = current.content.activeCell;
                if (activeCell) {
                    activeCell.model.value.text = "# this code is generated by the Low Code Assistant toolbar button\nimport low_code_assistant as lca\nlca.init()";
                    NotebookActions.runAndAdvance(current.content, current.sessionContext);
                }
            }
        });
        panel.toolbar.insertAfter('cellType', 'lcaInit', button);
        return button;
    }
}
const plugin = {
    id: 'lcalabextension:plugin',
    autoStart: true,
    requires: [INotebookTracker],
    activate: (app, NotebookTracker) => {
        initMixpanel();
        const getNotebookId = (notebookChildElement) => {
            const target = NotebookTracker.find(notebook => notebook.node.contains(notebookChildElement));
            return target === null || target === void 0 ? void 0 : target.id;
        };
        const getCells = (notebookId) => {
            const notebook = NotebookTracker.find(notebook => notebook.id === notebookId);
            if (notebook) {
                return notebook.content.widgets.map(cell => cell);
            }
        };
        const changeCellType = (notebookId, cell, cellType) => {
            var _a;
            const notebook = NotebookTracker.find(notebook => notebook.id === notebookId);
            const cellIndex = notebook === null || notebook === void 0 ? void 0 : notebook.content.widgets.indexOf(cell);
            if (notebook && notebook.model && cellIndex !== undefined) {
                const cellData = cell.model.toJSON();
                let newCell = null;
                if (cellType === 'code') {
                    newCell = notebook.model.contentFactory.createCodeCell({ cell: cellData });
                }
                else if (cellType === 'markdown') {
                    newCell = notebook.model.contentFactory.createMarkdownCell({ cell: cellData });
                }
                if (newCell !== null) {
                    (_a = notebook.model) === null || _a === void 0 ? void 0 : _a.cells.set(cellIndex, newCell);
                    return newCell;
                }
            }
        };
        const runAndAdvance = (notebookId) => {
            const notebook = NotebookTracker.find(notebook => notebook.id === notebookId);
            if (notebook) {
                NotebookActions.runAndAdvance(notebook.content, notebook.sessionContext);
            }
        };
        const getNotebookPath = (notebookId) => {
            const notebook = NotebookTracker.find(notebook => notebook.id === notebookId);
            if (notebook) {
                return notebook.context.path;
            }
        };
        const saveNotebook = async (notebookId) => {
            const notebook = NotebookTracker.find(notebook => notebook.id === notebookId);
            if (notebook) {
                try {
                    await notebook.context.save();
                    return true;
                }
                catch (e) { }
            }
            return false;
        };
        const assistants = {}; // maps from notebookId to assistant vue object
        const LCA = { frontend: 'lab', toArray, toObject, getNotebookId, getCells, assistants, changeCellType, runAndAdvance, getNotebookPath, saveNotebook };
        window["LCA"] = LCA;
        console.log('JupyterLab extension lcalabextension is activated!', NotebookTracker);
        const { commands, docRegistry } = app;
        if (NotebookTracker) {
            const lcaButton = new LCAInitButton(commands, NotebookTracker);
            docRegistry.addWidgetExtension('Notebook', lcaButton);
        }
    }
};
export default plugin;
